from __future__ import unicode_literals
import time
from .constants import LOGGER, *
from .roonapisocket import RoonApiWebSocket
from .discovery  import RoonDiscovery
import threading

class RoonApi():
    _roonsocket = None
    _roondiscovery = None
    _host = None
    _port = None
    _token = None
    _exit = False
    _zones = {}
    _outputs = {}
    _source_controls_request_id = None
    _volume_controls_request_id = None
    _source_controls = {}
    _volume_controls = {}
    _state_callbacks = []
    ready = False

    @property
    def token(self):
        ''' the authentication key that was retrieved from the registration with Roon'''
        return self._token

    @property
    def zones(self):
        ''' all zones, returned as dict'''
        return self._zones

    @property
    def outputs(self):
        ''' all outputs, returned as dict'''
        return self._outputs

    def zone_by_name(self, zone_name):
        ''' get zone details by name'''
        for zone in self.zones.values():
            if zone["display_name"] == zone_name:
                return zone
        return None

    def output_by_name(self, output_name):
        ''' get the output details from the name'''
        for output in self.outputs.values():
            if output["display_name"] == output_name:
                return output
        return None

    def zone_by_output_id(self, output_id):
        ''' get the zone details by output id'''
        for zone in self.zones.values():
            for output in zone["outputs"]:
                if output["output_id"] == output_id:
                    return zone
        return None

    def zone_by_output_name(self, output_name):
        '''
            get the zone details by an output name
            params:
                output_name: the name of the output
            returns: full zone details (dict)
        '''
        for zone in self.zones.values():
            for output in zone["outputs"]:
                if output["display_name"] == output_name:
                    return zone
        return None

    def get_image(self, image_key, scale="fit", width=500, height=500):
        '''
            get the image url for the specified image key
            params:
                image_key: the key for the image as retrieved in other api calls
                scale: optional (value of fit, fill or stretch)
                width: the width of the image (required if scale is specified)
                height: the height of the image (required if scale is set)
            returns: string with the full url to the image
        '''
        return "http://%s:%s/api/image/%s?scale=%s&width=%s&height=%s" %(self._host, self._port, image_key, scale, width, height)

    def playback_control(self, zone_or_output_id, control="play"):
        '''
            send player command to the specified zone
            params:
                zone_or_output_id: the id of the zone or output
                control:
                     * "play" - If paused or stopped, start playback
                     * "pause" - If playing or loading, pause playback
                     * "playpause" - If paused or stopped, start playback. If playing or loading, pause playback.
                     * "stop" - Stop playback and release the audio device immediately
                     * "previous" - Go to the start of the current track, or to the previous track
                     * "next" - Advance to the next track
        '''
        data = {
                   "zone_or_output_id": zone_or_output_id,
                   "control":           control
                }
        return self._request(ServiceTransport + "/control", data)

    def standby(self, output_id, control_key=None):
        '''
            send standby command to the specified output
            params:
                output_id: the id of the output to put in standby
                control_key: The control_key that identifies the source_control that is to be put into standby.
                             If omitted, then all source controls on this output that support standby will be put into standby.
        '''
        data = {  "output_id": output_id, "control_key": control_key }
        return self._request(ServiceTransport + "/standby", data)

    def convenience_switch(self, output_id, control_key=None):
        '''
            Convenience switch an output, taking it out of standby if needed.
            params:
                output_id: the id of the output that should be convenience-switched.
                control_key: The control_key that identifies the source_control that is to be switched.
                             If omitted, then all controls on this output will be convenience switched.
        '''
        data = {  "output_id": output_id, "control_key": control_key }
        return self._request(ServiceTransport + "/convenience_switch", data)

    def mute(self, output_id, mute=True):
        '''
            Mute/unmute an output.
            params:
                output_id: the id of the output that should be muted/unmuted
                mute: bool if the output should be muted. Will unmute if set to False
        '''
        how = "mute" if mute else "unmute"
        data = {  "output_id": output_id, "how": how }
        return self._request(ServiceTransport + "/mute", data)

    def change_volume(self, output_id, value, method="absolute"):
        '''
            Change the volume of an output. For convenience you can always just give te new volume level as percentage
            params:
                output_id: the id of the output
                value: The new volume value, or the increment value or step (as percentage)
                method: How to interpret the volume ('absolute'|'relative'|'relative_step')
        '''
        if not "volume" in self._outputs[output_id]:
            LOGGER.info("This endpoint has fixed volume.")
            return None
        # Home assistant was catching this - so catch here to try and diagnose what needs to be checked.
        try:
            if method == "absolute":
                if self._outputs[output_id]["volume"]["type"] == "db":
                    value = int((float(value) / 100) * 80) - 80
            data = {  "output_id": output_id, "how": method, "value": value }
            return self._request(ServiceTransport + "/change_volume", data)
        except Exception as exc:  # pylint: disable=broad-except
            LOGGER.error("set_volume_level failed for entity %s \n %s.", str(exc))
            return None


    def seek(self, zone_or_output_id, seconds, method="absolute"):
        '''
            Seek to a time position within the now playing media
            params:
                zone_or_output_id: the id of the zone or output
                seconds: The target seek position
                method: How to interpret the target seek position ('absolute'|'relative')
        '''
        data = {  "zone_or_output_id": zone_or_output_id, "how": method, "seconds": seconds }
        return self._request(ServiceTransport + "/seek", data)

    def shuffle(self, zone_or_output_id, shuffle=True):
        '''
            Enable/disable shuffle
            params:
                zone_or_output_id: the id of the output or zone
                shuffle: bool if shuffle should be enabled. False will disable shuffle
        '''
        data = {  "zone_or_output_id": zone_or_output_id, "shuffle": shuffle }
        return self._request(ServiceTransport + "/change_settings", data)

    def repeat(self, zone_or_output_id, repeat=True):
        '''
            Enable/disable repeat (loop mode)
            params:
                zone_or_output_id: the id of the output or zone
                repeat: bool if repeat should be enabled. False will disable shuffle
        '''
        loop = "loop" if repeat else "disabled"
        data = {  "zone_or_output_id": zone_or_output_id, "loop": loop }
        return self._request(ServiceTransport + "/change_settings", data)

    def transfer_zone(self, from_zone_or_output_id, to_zone_or_output_id):
        '''
            Transfer the current queue from one zone to another
            params:
                from_zone_or_output_id - The source zone or output
                to_zone_or_output_id - The destination zone or output
        '''
        data = { "from_zone_or_output_id": from_zone_or_output_id,
                    "to_zone_or_output_id": to_zone_or_output_id }
        return self._request(ServiceTransport + "/transfer_zone", data)

    def group_outputs(self, output_ids):
        '''
            Create a group of synchronized audio outputs
            params:
                output_ids - The outputs to group. The first output's zone's queue is preserved.
        '''
        data = { "output_ids": output_ids }
        return self._request(ServiceTransport + "/group_outputs", data)

    def ungroup_outputs(self, output_ids):
        '''
            Ungroup outputs previous grouped
            params:
                output_ids - The outputs to ungroup.
        '''
        data = { "output_ids": output_ids }
        return self._request(ServiceTransport + "/ungroup_outputs", data)

    def register_source_control(self, control_key, display_name, callback, initial_state="selected", supports_standby=True):
        ''' register a new source control on the api'''
        if control_key in self._source_controls:
            LOGGER.error("source_control %s is already registered!" % control_key)
            return
        control_data = {
                    "display_name": display_name,
                    "supports_standby": supports_standby,
                    "status": initial_state,
                    "control_key": control_key
                    }
        self._source_controls[control_key] = (callback, control_data)
        if self._source_controls_request_id:
            data = {"controls_added":[ control_data ]}
            self._roonsocket.send_continue(self._source_controls_request_id, data)

    def update_source_control(self, control_key, new_state):
        ''' update an existing source control on the api'''
        if control_key not in self._source_controls:
            LOGGER.warning("source_control %s is not (yet) registered!" % control_key)
            return
        if not self._source_controls_request_id:
            LOGGER.warning("Not yet registered, can not update source control")
            return False
        self._source_controls[control_key][1]["status"] = new_state
        data = {"controls_changed": [ self._source_controls[control_key][1] ] }
        self._roonsocket.send_continue(self._source_controls_request_id, data)

    def register_volume_control(self, control_key, display_name, callback, initial_volume=0, volume_type="number", volume_step=2, volume_min=0, volume_max=100, is_muted=False):
        ''' register a new volume control on the api'''
        if control_key in self._volume_controls:
            LOGGER.error("source_control %s is already registered!" % control_key)
            return
        control_data = {
                    "display_name": display_name,
                    "volume_type": volume_type,
                    "volume_min": volume_min,
                    "volume_max": volume_max,
                    "volume_value": initial_volume,
                    "volume_step": volume_step,
                    "is_muted": is_muted,
                    "control_key": control_key
                    }
        self._volume_controls[control_key] = (callback, control_data)
        if self._volume_controls_request_id:
            data = {"controls_added":[ control_data ]}
            self._roonsocket.send_continue(self._volume_controls_request_id, data)

    def update_volume_control(self, control_key, volume=None, mute=None):
        ''' update an existing volume control, report its state to Roon '''
        if control_key not in self._volume_controls:
            LOGGER.warning("volume_control %s is not (yet) registered!" % control_key)
            return
        if not self._volume_controls_request_id:
            LOGGER.warning("Not yet registered, can not update volume control")
            return False
        if volume != None:
            self._volume_controls[control_key][1]["volume_value"] = volume
        if mute != None:
            self._volume_controls[control_key][1]["is_muted"] = mute
        data = {"controls_changed": [ self._volume_controls[control_key][1] ] }
        self._roonsocket.send_continue(self._volume_controls_request_id, data)

    def register_state_callback(self, callback, event_filter=None, id_filter=None):
        '''
            register a callback to be informed about changes to zones or outputs
            params:
                callback: method to be called when state changes occur, it will be passed an event param as string and a list of changed objects
                          callback will be called with params:
                          - event: string with name of the event ("zones_changed", "zones_seek_changed", "outputs_changed")
                          - a list with the zone or output id's that changed
                event_filter: only callback if the event is in this list
                id_filter: one or more zone or output id's or names to filter on (list or string)
        '''
        if not event_filter:
            event_filter = []
        elif not isinstance(event_filter, list):
            event_filter = [event_filter]
        if not id_filter:
            id_filter = []
        elif not isinstance(id_filter, list):
            id_filter = [id_filter]
        self._state_callbacks.append( (callback, event_filter, id_filter) )

    def register_queue_callback(self, callback, zone_or_output_id=""):
        '''
            subscribe to queue change events
            callback: function which will be called with the updated data (provided as dict object
            zone_or_output_id: If provided, only listen for updates for this zone or output
        '''
        if zone_or_output_id:
            opt_data = {"zone_or_output_id": zone_or_output_id}
        else:
            opt_data = None
        self._roonsocket.subscribe(ServiceTransport, "queue", callback, opt_data)

    def browse_browse(self, opts):
        '''
            undocumented and complex browse call on the roon api
            reference: https://github.com/RoonLabs/node-roon-api-browse/blob/master/lib.js
        '''
        return self._request(ServiceBrowse + "/browse", opts)

    def browse_load(self, opts):
        '''
            undocumented and complex browse call on the roon api
            reference: https://github.com/RoonLabs/node-roon-api-browse/blob/master/lib.js
        '''
        return self._request(ServiceBrowse + "/load", opts)

    def browse_pop_all(self, opts):
        '''
            undocumented and complex browse call on the roon api
            reference: https://github.com/RoonLabs/node-roon-api-browse/blob/master/lib.js
        '''
        return self._request(ServiceBrowse + "/pop_all", opts)

    def browse_pop(self, opts):
        '''
            undocumented and complex browse call on the roon api
            reference: https://github.com/RoonLabs/node-roon-api-browse/blob/master/lib.js
        '''
        return self._request(ServiceBrowse + "/pop", opts)

    def browse_by_path(self, search_paths, zone_or_output_id="", offset=0, search_input=None):
        '''
            workaround to browse content by specifying the path to the content
            params:
                search_paths: a list of names to look for in the hierarchie.
                              e.g. ["Playlists", "My Favourite Playlist"]
                zone_or_output_id: id of a zone or output on which behalf the search is performed.
                              can be ommitted for browsing but required for actions (play etc.)
                offset: a list will only return 100 items, to get more use the offset
            returns: a list of items (if found) as returned by the browse function
        '''
        opts = {"hierarchy": "browse", "pop_all": True}
        if zone_or_output_id:
            opts["zone_or_output_id"] = zone_or_output_id
        # go to first level (home)
        result = self.browse_browse(opts)
        if not result:
            return None
        # items at first level (mainmenu items)
        result = self.browse_load(opts)
        opts["pop_all"] = False
        for search_for in search_paths:
            if not result or not "items" in result:
                break
            for item in result["items"]:
                if item["title"] == search_for or search_input and item.get("input_prompt"):
                    opts["item_key"] = item["item_key"]
                    if item.get("input_prompt"):
                        opts["input"] = search_input
                    result = self.browse_browse(opts)
                    if result and  "list" in result and result["list"]["count"] > 100:
                        opts["offset"] = offset
                        opts["set_display_offset"] = offset
                    result = self.browse_load(opts)
        return result

    def playlists(self, offset=0):
        ''' return the list of playlists'''
        return self.browse_by_path(["Playlists"], offset=offset)

    def internet_radio(self, offset=0):
        ''' return the list of internet radio stations'''
        return self.browse_by_path(["Internet Radio"], offset=offset)

    def artists(self, offset=0):
        '''return the list of artists in the library'''
        return self.browse_by_path(["Library", "Artists"], offset=offset)

    def albums(self, offset=0):
        '''return the list of albums in the library'''
        return self.browse_by_path(["Library", "Albums"], offset=offset)

    def tracks(self, offset=0):
        '''return the list of tracks in the library'''
        return self.browse_by_path(["Library", "Tracks"], offset=offset)

    def tags(self, offset=0):
        '''return the list of tags in the library'''
        return self.browse_by_path(["Library", "Tags"], offset=offset)

    def genres(self, subgenres_for="", offset=0):
        '''return the list of genres in the library'''
        return self.browse_by_path(["Genres", subgenres_for], offset=offset)

    def play_playlist(self, zone_or_output_id, playlist_title, shuffle=False):
        ''' play playlist by name on the specified zone'''
        play_action = "Shuffle" if shuffle else "Play Now"
        return self.browse_by_path(["Playlists", playlist_title, "Play Playlist", play_action], zone_or_output_id)

    def queue_playlist(self, zone_or_output_id, playlist_title):
        ''' queue playlist by name on the specified zone'''
        return self.browse_by_path(["Playlists", playlist_title, "Play Playlist", "Queue"], zone_or_output_id)

    def play_radio(self, zone_or_output_id, radio_title):
        ''' play internet radio by name on the specified zone'''
        return self.browse_by_path(["Internet Radio", radio_title, "Play Radio", "Play Now"], zone_or_output_id)

    def play_genre(self, zone_or_output_id, genre_name, subgenre="", shuffle=False):
        '''play specified genre on the specified zone'''
        action = "Shuffle" if shuffle else "Play Genre"
        if subgenre:
            return self.browse_by_path(["Genres", genre_name, subgenre, "Play Genre", action], zone_or_output_id)
        else:
            return self.browse_by_path(["Genres", genre_name, "Play Genre", action], zone_or_output_id)

    def play_id(self, zone_or_output_id, media_id):
        '''Play based on the media_id from the browse api.'''
        opts = {
            "zone_or_output_id": zone_or_output_id,
            "item_key": media_id,
            "hierarchy": "browse",
        }
        header_result = self.browse_browse(opts)
        # For Internet radio the above load starts play - so catch this and return
        try:
            if header_result["list"]["level"] == 0:
                LOGGER.info("Initial load started playback")
                return
        except NameError:
            pass

        if header_result is None:
            LOGGER.error(
                "Playback requested of unsupported id: %s",
                media_id,
            )
            return

        result = self.browse_load(opts)

        first_item = result["items"][0]
        hint = first_item["hint"]
        LOGGER.warning(first_item)
        if not (hint == "action" or hint == "action_list"):
            LOGGER.error(
                "Playback requested but item is a list, not a playable action or action_list id: %s",
                media_id,
            )
            return

        if hint == "action_list":
            opts["item_key"] = first_item["item_key"]
            result = self.browse_browse(opts)
            if result is None:
                LOGGER.error(
                    "Playback requested of unsupported id: %s",
                    media_id,
                )
                return
            result = self.browse_load(opts)
            first_item = result["items"][0]
            hint = first_item["hint"]

        if hint != "action":
            LOGGER.error(
                "Playback requested but item does not have a playable action id: %s, %s",
                media_id, header_result
            )
            return

        play_action = result["items"][0]
        hint = play_action["hint"]
        LOGGER.info("'%s' for '%s')", play_action["title"], header_result)
        opts["item_key"] = play_action["item_key"]
        self.browse_browse(opts)
        if result is None:
            LOGGER.error(
                "Playback requested of unsupported id: %s",
                media_id,
            )

    def search_artists(self, search_input):
        ''' search for artists by name'''
        return self.browse_by_path(["Library", "Search", "Artists"], search_input=search_input)


    ############# private methods ##################


    def __init__(self, appinfo, token=None, host=None, port=9100, blocking_init=True):
        '''
            Set up the connection with Roon
            appinfo: a dict of the required information about the app that should be connected to the api
            token: used for presistant storage of the auth token, will be set to token attribute if retrieved. You should handle saving of the key yourself
            host: optional the ip or hostname of the Roon server, will be auto discovered if ommitted
            port: optional the http port of the Roon websockets api. Should be default of 9100
            blocking_init: By default the init will halt untill the socket is connected and the app is authenticated,
                           if you set bool to False the init will continue but you will only receive data once the connection is fully initialized.
                           The latter is preferred if you're (only) using the callbacks
        '''
        self._appinfo = appinfo
        self._token = token
        if not appinfo or not isinstance(appinfo, dict):
            raise("appinfo missing or in incorrect format!")

        if host and port:
            self._server_discovered(host, port)
        else:
            self._roondiscovery = RoonDiscovery(self._server_discovered)
            self._roondiscovery.start()
        # block untill we're ready
        if blocking_init:
            while not self.ready and not self._exit:
                time.sleep(1)
        # start socket watcher
        th = threading.Thread(target=self._socket_watcher)
        th.daemon = True
        th.start()

    def __exit__(self, type, value, tb):
        self.stop()

    def __enter__(self):
        return self

    def stop(self):
        self._exit = True
        if self._roondiscovery:
            self._roondiscovery.stop()
        if self._roonsocket:
            self._roonsocket.stop()

    def _server_discovered(self, host, port):
        ''' called when the roon server is (auto) discovered on the network'''
        LOGGER.info("Connecting to Roon server %s:%s" % (host, port))
        ws_address = "ws://%s:%s/api" %(host, port)
        self._host = host
        self._port = port
        self._roonsocket = RoonApiWebSocket(ws_address)
        self._roonsocket.source_controls_callback = self._on_source_control_request
        self._roonsocket.volume_controls_callback = self._on_volume_control_request
        self._roonsocket.connected_callback = self._socket_connected
        self._roonsocket.registered_calback = self._server_registered
        self._roonsocket.start()

    def _socket_connected(self):
        ''' the websocket connection is connected successfully'''
        LOGGER.info("Connection with roon websockets (re)created.")
        self.ready = False
        self._source_controls_request_id = None
        self._volume_controls_request_id = None
        # authenticate / register
        # warning: at first launch the user has to approve the app in the Roon settings.
        appinfo = self._appinfo.copy()
        appinfo["required_services"] = [ServiceTransport, ServiceBrowse]
        appinfo["provided_services"] = [ControlVolume, ControlSource]
        if self._token:
            appinfo["token"] = self._token
        if not self._token:
            LOGGER.info("The application should be approved within Roon's settings.")
        else:
            LOGGER.info("Registering the app with Roon...")
        self._roonsocket.send_request(ServiceRegistry + "/register", appinfo)

    def _server_registered(self, reginfo):
        LOGGER.info("Registered to Roon server %s" % reginfo["display_name"])
        LOGGER.debug(reginfo)
        self._token = reginfo["token"]
        # fill zones and outputs dicts one time so the data is available right away
        if not self._zones:
            self._zones = self._get_zones()
        if not self._outputs:
            self._outputs = self._get_outputs()
        # subscribe to state change events
        self._roonsocket.subscribe(ServiceTransport, "zones", self._on_state_change)
        self._roonsocket.subscribe(ServiceTransport, "outputs", self._on_state_change)
        # set flag that we're fully initialized (used for blocking init)
        self.ready = True

    def _on_state_change(self, msg):
        ''' process messages we receive from the roon websocket into a more usable format'''
        events = []
        if not msg or not isinstance(msg, dict):
            return
        for state_key, state_values in msg.items():
            changed_ids = []
            filter_keys = []
            if state_key in ["zones_seek_changed", "zones_changed", "zones_added", "zones"]:
                for zone in state_values:
                    if zone["zone_id"] in self._zones:
                        self._zones[zone["zone_id"]].update(zone)
                    else:
                        self._zones[zone["zone_id"]] = zone
                    changed_ids.append(zone["zone_id"])
                    if "display_name" in zone:
                        filter_keys.append(zone["display_name"])
                    if "outputs" in zone:
                        for output in zone["outputs"]:
                            filter_keys.append(output["output_id"])
                            filter_keys.append(output["display_name"])
                event = "zones_seek_changed" if state_key == "zones_seek_changed" else "zones_changed"
                events.append((event, changed_ids, filter_keys))
            elif state_key in ["outputs_changed", "outputs_added", "outputs"]:
                for output in state_values:
                    if output["output_id"] in self._outputs:
                        self._outputs[output["output_id"]].update(output)
                    else:
                        self._outputs[output["output_id"]] = output
                    changed_ids.append(output["output_id"])
                    filter_keys.append(output["display_name"])
                    filter_keys.append(output["zone_id"])
                event = "outputs_changed"
                events.append((event, changed_ids, filter_keys))
            elif state_key == "zones_removed":
                for item in state_values:
                    del self._zones[item]
            elif state_key == "outputs_removed":
                for item in state_values:
                    del self._outputs[item]
            else:
                LOGGER.warning("unknown state change: %s" % msg)
        for event, changed_ids, filter_keys in events:
            filter_keys.extend(changed_ids)
            for item in self._state_callbacks:
                callback = item[0]
                event_filter = item[1]
                id_filter = item[2]
                if event_filter and (event not in event_filter):
                    continue
                if id_filter and set(id_filter).isdisjoint(filter_keys):
                    continue
                try:
                    callback(event, changed_ids)
                except Exception:
                    LOGGER.exception("Error while executing callback!")

    def _get_outputs(self):
        outputs = {}
        data = self._request(ServiceTransport + "/get_outputs")
        if data and "outputs" in data:
            for output in data["outputs"]:
                outputs[output["output_id"]] = output
        return outputs

    def _get_zones(self):
        zones = {}
        data = self._request(ServiceTransport + "/get_zones")
        if data and "zones" in data:
            for zone in data["zones"]:
                zones[zone["zone_id"]] = zone
        return zones

    def _request(self, command, data=None):
        ''' send command and wait for result '''
        if not self._roonsocket:
            retries = 20
            while (not self.ready or not self._roonsocket) and retries:
                retries -= 1
                time.sleep(0.2)
            if not self.ready or not self._roonsocket:
                LOGGER.warning("socket is not yet ready")
                if not self._roonsocket:
                    return None
        request_id = self._roonsocket.send_request(command, data)
        result = None
        retries = 50
        while retries:
            result = self._roonsocket.results.get(request_id)
            if result:
                break
            else:
                retries -= 1
                time.sleep(0.05)
        try:
            del self._roonsocket.results[request_id]
        except KeyError:
            pass
        return result

    def _on_source_control_request(self, event, request_id, data):
        ''' got request from roon server for a source control registered on this endpoint'''
        if event == "subscribe_controls":
            LOGGER.debug("found subscription ID for source controls: %s " % request_id)
            self._roonsocket.send_continue(request_id, {"controls_added": []})
            # send all source controls already registered (handle connection loss)
            controls = []
            for callback, control_data in self._source_controls.values():
                controls.append(control_data)
            self._roonsocket.send_continue(request_id, { "controls_added":controls })
            self._source_controls_request_id = request_id
        elif data and data.get("control_key"):
            control_key = data["control_key"]
            try:
                # launch callback
                self._roonsocket.send_complete(request_id, "Success")
                self._source_controls[control_key][0](control_key, event)
            except Exception:
                LOGGER.exception("Error in source_control callback")
                self._roonsocket.send_complete(request_id, "Error")

    def _on_volume_control_request(self, event, request_id, data):
        ''' got request from roon server for a volume control registered on this endpoint'''
        if event == "subscribe_controls":
            LOGGER.debug("found subscription ID for volume controls: %s " % request_id)
            # send all volume controls already registered (handle connection loss)
            controls = []
            for callback, control_data in self._volume_controls.values():
                controls.append(control_data)
            self._roonsocket.send_continue(request_id, { "controls_added":controls })
            self._source_controls_request_id = request_id
            self._volume_controls_request_id = request_id
        elif data and data.get("control_key"):
            control_key = data["control_key"]
            if event == "set_volume" and data["mode"] == "absolute":
                value = data["value"]
            elif event == "set_volume" and data["mode"] == "relative":
                value = self._volume_controls[control_key][0]["volume_value"] + data["value"]
            elif event == "set_volume" and data["mode"] == "relative_step":
                value = self._volume_controls[control_key][0]["volume_value"] + (data["value"] * data["volume_step"])
            elif event == "set_mute":
                value = data["mode"] == "on"
            else:
                return
            try:
                self._roonsocket.send_complete(request_id, "Success")
                self._volume_controls[control_key][0](control_key, event, value)
            except Exception:
                LOGGER.exception("Error in volume_control callback")
                self._roonsocket.send_complete(request_id, "Error")

    def _socket_watcher(self):
        ''' monitor the connection state of the socket and reconnect if needed'''
        while not self._exit:
            if self._roonsocket and self._roonsocket.failed_state:
                LOGGER.warning("Socket connection lost! Will try to reconnect in 20s")
                count = 0
                while not self._exit and count < 21:
                    count += 1
                    time.sleep(1)
                if not self._exit:
                    self._server_discovered(self._host, self._port)
            time.sleep(2)



from __future__ import unicode_literals
import logging
ServiceRegistry     = "com.roonlabs.registry:1"
ServiceTransport    = "com.roonlabs.transport:2"
ServiceStatus       = "com.roonlabs.status:1"
ServicePairing      = "com.roonlabs.pairing:1"
ServicePing         = "com.roonlabs.ping:1"
ServiceImage        = "com.roonlabs.image:1"
ServiceBrowse       = "com.roonlabs.browse:1"
ServiceSettings     = "com.roonlabs.settings:1"
ControlVolume       = "com.roonlabs.volumecontrol:1"
ControlSource       = "com.roonlabs.sourcecontrol:1"

MessageRequest      = "REQUEST"
MessageComplete     = "COMPLETE"
MessageContinue     = "CONTINUE"


logformat = logging.Formatter('%(asctime)-15s %(levelname)-5s  %(module)s -- %(message)s')
LOGGER = logging.getLogger("roonapi")
consolehandler = logging.StreamHandler()
consolehandler.setFormatter(logformat)
LOGGER.addHandler(consolehandler)
LOGGER.setLevel(logging.INFO)


from __future__ import unicode_literals
import websocket
import threading
import time
from .constants import *

try:
    import simplejson as json
except ImportError:
    import json

try:
    import thread
except ImportError:
    import _thread as thread


class RoonApiWebSocket(threading.Thread):
    _socket = None
    _results = {}
    _requestid = 10 # initial request_id of 10 to prevent confusion with the requests that are sent by the server at initialization
    _subkey = 0
    _exit = False
    _subscriptions = {}
    source_controls_callback = None
    volume_controls_callback = None
    connected_callback = None
    registered_calback = None
    connected = False
    failed_state = False

    @property
    def results(self):
        return self._results

    def run(self):
        self._socket.run_forever(ping_interval=10)
        if not self._exit:
            LOGGER.warning("Session unexpectedly disconnected!")
            self._exit = True
            self.failed_state = True
        else:
            LOGGER.debug("socket connection closed")

    def stop(self):
        self._exit = True
        subscriptions = []
        for key, value in self._subscriptions.items():
            subscriptions.append( (value["service"], value["endpoint"]) )
        for service, endpoint in subscriptions:
            self.unsubscribe(service, subscriptions)
        self._socket.close()

    def __init__(self, host):
        self._socket = websocket.WebSocketApp(host, 
                on_message=self.on_message, 
                on_error=self.on_error, 
                on_open=self.on_open, 
                on_close=self.on_close)
        threading.Thread.__init__(self)
        self.daemon = True
        
    def subscribe(self, service, endpoint, callback, opt_data=None):
        '''subscribe to events'''
        subkey = self._subkey
        self._subkey += 1
        data = {u"subscription_key": subkey}
        if opt_data and isinstance(opt_data, dict):
            data.update(opt_data)
        request_id = self.send_request(service + "/subscribe_" + endpoint, data)
        self._subscriptions[request_id] = {
            "service": service,
            "endpoint": endpoint,
            "request_id": request_id,
            "subkey": subkey,
            "callback": callback
        }

    def unsubscribe(self, service, endpoint):
        '''subscribe to events'''
        matches = []
        for key, value in self._subscriptions.items():
            if value["service"] == service and value["endpoint"] == endpoint:
                matches.append((key, value["subkey"]))
        for item in matches:
            self.send_request(service + "/unsubscribe_" + endpoint, {"subscription_key": item[1]})
            del self._subscriptions[item[0]]

    def on_message(self, ws, message=None):
        if not message:
            message = ws # compatability fix because of change in websocket-client v0.49
        try:
            message = message.decode('utf-8')
            lines = message.split("\n")
            header = lines[0]
            body = ""
            if "Content-Type:" in message:
                content_type = lines[1].split("Content-Type: ")[1]
                request_id = int(lines[2].split("Request-Id: ")[1])
                content_length = int(lines[3].split("Content-Length: ")[1])
                body = "".join(lines[5:])
            elif "Logging:" in message:
                request_id = int(lines[2].split("Request-Id: ")[1])
            else:
                request_id = int(lines[1].split("Request-Id: ")[1])
                body = header
            if body and "{" in body:
                body = json.loads(body)
            # handle message
            if ControlSource in header:
                # incoming message for source_control endpoint
                event = header.split("/")[-1]
                if self.source_controls_callback:
                   self.source_controls_callback(event, request_id, body)
            elif ControlVolume in header:
                # incoming message for volume_control endpoint
                event = header.split("/")[-1]
                if self.volume_controls_callback:
                    self.volume_controls_callback(event, request_id, body)
            elif ServicePing in header:
                # reply to incoming ping from server
                self.send_complete(request_id, "Success")
            elif "Registered" in header:
                if self.registered_calback:
                    self.registered_calback(body)
            elif request_id in self._subscriptions:
                # this is callback for one of our subscriptions
                self._subscriptions[request_id]["callback"](body)
            else:
                # this is just a result for one of our requests
                self._results[request_id] = body
        except Exception as exc:
            LOGGER.exception("Error while parsing message")
            LOGGER.debug(message)

    def on_error(self, ws, error=None):
        if not error:
            error = ws # compatability fix because of change in websocket-client v0.49
        LOGGER.error(error)

    def on_close(self, ws=None):
        LOGGER.info("session closed")
        self.connected = False
        self._requestid = 10
        self._subkey = 0
        self._subscriptions = {}
        
    def on_open(self, ws=None):
        LOGGER.debug('Opened Websocket connection to the server...')
        self.connected = True
        if self.connected_callback:
            thread.start_new_thread(self.connected_callback, ())
        
    def send_continue(self, request_id, body):
        if not self.connected:
            LOGGER.error("Connection is not (yet) ready!")
            return False
        body = json.dumps(body)
        msg = u"MOO/1 CONTINUE Changed\nRequest-Id: %s\nContent-Length: %s\nContent-Type: application/json\n\n%s" %(request_id, len(body), body)
        try:
            msg = bytes(msg) # py2
        except TypeError:
            msg = bytes(msg, 'utf-8') # py3
        self._socket.send(msg, 0x2)

    def send_complete(self, request_id, name, body=""):
        if not self.connected:
            LOGGER.error("Connection is not (yet) ready!")
            return False
        msg = u"MOO/1 COMPLETE %s\nRequest-Id: %s" %(name, request_id)
        if body:
            body = json.dumps(body)
            msg += u"\nContent-Length: %s\nContent-Type: application/json\n\n%s" %(len(body), body)
        else:
            msg += "\n\n"
        try:
            msg = bytes(msg) # py2
        except TypeError:
            msg = bytes(msg, 'utf-8') # py3
        self._socket.send(msg, 0x2)

    def send_request(self, command, body=None, content_type="application/json", header_type="REQUEST"):
        if not self.connected:
            LOGGER.error("Connection is not (yet) ready!")
            return False
        request_id = self._requestid
        self._requestid += 1
        self._results[request_id] = None
        if body == None:
            msg = u"MOO/1 REQUEST %s\nRequest-Id: %s\n\n" % (command, request_id)
        else:
            body = json.dumps(body)
            msg = u"MOO/1 REQUEST %s\nRequest-Id: %s\nContent-Length: %s\nContent-Type: %s\n\n%s" %(command, request_id, len(body), content_type, body)
        try:
            msg = bytes(msg) # py2
        except TypeError:
            msg = bytes(msg, 'utf-8') # py3
        self._socket.send(msg, 0x2)
        return request_id



"""
Module defining a class to discover Roon servers.

If multiple servers are available on the network, the first to be discovered
is selected. This may not be the one you have enabled the plugin for.
"""

import threading
import socket
import os.path
from .soodmessage import SOODMessage, FormatException


class RoonDiscovery(threading.Thread):
    """Class to discover Roon Servers connected in the network."""
    _exit = threading.Event()
    _discovered_callback = None

    def __init__(self, callback):
        self._discovered_callback = callback
        threading.Thread.__init__(self)
        self.daemon = True

    def run(self):
        ''' run discovery until server found '''
        while not self._exit.isSet():
            host, port = self.first()
            if host:
                self._discovered_callback(host, port)
                self.stop()

    def stop(self):
        self._exit.set()

    def all(self):
        """Scan and return all found entries as a list. Each server is a tuple of host,port."""
        self._discover(first_only=False)

    def first(self):
        ''' returns first server that is found'''
        all_servers = self._discover(first_only=True)
        return all_servers[0] if all_servers else (None, None)

    def _discover(self, first_only=False):
        """update the server entry with details"""
        this_dir = os.path.dirname(os.path.abspath(__file__))
        sood_file = os.path.join(this_dir, ".soodmsg")
        with open(sood_file) as sood_query_file:
            msg = sood_query_file.read()
        msg = msg.encode()
        entries = []
        with socket.socket(socket.AF_INET, socket.SOCK_DGRAM) as sock:
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
            sock.settimeout(5)
            sock.bind(('', 0))
            sock.sendto(msg, ('<broadcast>', 9003))
            while not self._exit.isSet():
                try:
                    data, server = sock.recvfrom(1024)
                    message = SOODMessage(data).as_dictionary
                    host = server[0]
                    port = message["properties"]["http_port"]
                    entries.append((host, port))
                    if first_only:
                        # we're only interested in the first server found
                        break
                except socket.timeout:
                    print("timeout")
                    break
                except FormatException as format_exception:
                    print(format_exception.message)
                    break
        return entries



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['roonapi']

package_data = \
{'': ['*']}

install_requires = \
['ifaddr>=0.1.0', 'requests>=2.0', 'six>=1.10.0']

setup_kwargs = {
    'name': 'roonapi',
    'version': '0.0.22.dev1',
    'description': 'Provides a python interface to interact with Roon',
    'long_description': '# python-roon\npython library to interface with the Roon API (www.roonlabs.com)\n\nFull documentation will follow asap\nSee the tests folder for some more code examples.\n\n\nSome example code:\n\n```\nfrom roon import RoonApi  \nappinfo = {\n        "extension_id": "python_roon_test",\n        "display_name": "Python library for Roon",\n        "display_version": "1.0.0",\n        "publisher": "marcelveldt",\n        "email": "mygreat@emailaddress.com"\n    }\n    \ntoken = open(\'mytokenfile\').read()\nroonapi = RoonApi(appinfo, token)\n\n# get all zones (as dict)\nprint(roonapi.zones)\n\n# get all outputs (as dict)\nprint(roonapi.outputs)\n\n# receive state updates in your callback\nroonapi.register_state_callback(my_state_callback)\n\n\n# save the token for next time\nwith open(\'mytokenfile\', \'w\') as f:\n    f.write(roonapi.token)\n',
    'author': 'Marcel van der Veldt',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/pavoni/pyroon',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



'''
SOOD messages consists of a header and a list of properties encoded as key/value pairs.
The header starts with "SOOD", followed by a byte with the value 2, followed by a byte representing the type.
The type is either 'Q' for "Query" or # 'R' for "Response.
The key/value pairs follows directly after the type.
The keys and values are prepended with one or two bytes with length info. One byte for keys,
two bytes for values.
In other words:
SOOD\x02<onelettertype><1bytelen><key><2bytelen><value><1bytelen><key><2bytelen><value>...

All lengths are big-endian.

The service_id always has the value 00720724-5143-4a9b-abac-0e50cba674bb

Query format:
SOOD\x02Q<1bytelen>query_service_id<2bytelen>00720724-5143-4a9b-abac-0e50cba674bb<1bytelen>_tid<2bytelen><the_tid>

Response format:
SOOD\x02R<1bytelen>name<2bytelen><the_name><1bytelen>display_version<2bytelen><the_version><1bytelen>unique_id<2bytelen><the_id><1bytelen>service_id<twobytelen>00720724-5143-4a9b-abac-0e50cba674bb<1bytelen>tcp_port<twobytelen><the_port><1bytelen>http_port<twobytelen><the_port><1bytelen>_tid<twobytelen>c64e3888-f2f2-4c4a-9f89-2093ae4217a6
'''


from enum import Enum, auto


class FormatException(Exception):
    """Exception to be raised on errors in a binary SOOD message"""
    def __init__(self, message):
        Exception.__init__()
        self.message = message


class SOODMessage:
    """Class for parsing SOOD messages."""
    __MESSAGE_PREFIX__ = b'SOOD\x02'

    class SOODMessageType(Enum):
        """Symbolic names for the message types"""
        QUERY = auto()
        RESPONSE = auto()

        def __repr__(self):
            return f"<{self.__class__.__name__}, {self.name}>"

    def __init__(self, message):
        if not message.startswith(self.__MESSAGE_PREFIX__):
            raise FormatException("Error in message header")
        self._message = message
        self._current_position = len(self.__MESSAGE_PREFIX__)

    def _parse_property(self, size_of_size):
        length = int.from_bytes(self._message[self._current_position:self._current_position+size_of_size], "big")
        self._current_position += size_of_size
        if self._current_position + length > len(self._message):
            return None
        part_string = self._message[self._current_position:self._current_position+length].decode()
        self._current_position += len(part_string)
        return part_string

    def _parse_properties(self):
        properties = {}
        while self._current_position < len(self._message):
            part_key = self._parse_property(1)
            if part_key is None:
                return None
            part_value = self._parse_property(2)
            if part_value is None:
                return None
            properties[part_key] = part_value
        return properties

    def _parse_type(self):
        type_letter = chr(self._message[self._current_position])
        self._current_position += 1
        if type_letter not in ["Q", "R"]:
            return None
        return self.SOODMessageType.QUERY if type_letter == "Q" else self.SOODMessageType.RESPONSE

    @property
    def as_dictionary(self):
        """Expose the message as a dictionary."""
        message_type = self._parse_type()
        if message_type is None:
            raise FormatException("Error in message type")

        message_properties = self._parse_properties()
        if message_properties is None:
            raise FormatException("Error in property")

        message = {
            "type": message_type,
            "properties": message_properties,
        }
        return message
